@@ -1,16 +1,64 @@
+pub(crate) use self::context_module::ContextModule;
 pub(crate) use self::context_selector::ContextSelector;
 pub(crate) use self::display::{Display, DisplayMatchArm};
 pub(crate) use self::error::{Error, ErrorSourceMatchArm};
 pub(crate) use self::error_compat::{ErrorCompat, ErrorCompatBacktraceMatchArm};
 
-pub mod context_selector {
-    use crate::{ContextSelectorKind, Field};
+pub mod context_module {
+    use crate::ModuleName;
+    use heck::ToSnakeCase;
     use proc_macro2::TokenStream;
-    use quote::{format_ident, quote, ToTokens};
+    use quote::{quote, ToTokens};
+    use syn::Ident;
+
+    #[derive(Copy, Clone)]
+    pub(crate) struct ContextModule<'a, T> {
+        pub container_name: &'a Ident,
+        pub module_name: &'a ModuleName,
+        pub visibility: Option<&'a dyn ToTokens>,
+        pub body: &'a T,
+    }
+
+    impl<'a, T> ToTokens for ContextModule<'a, T>
+    where
+        T: ToTokens,
+    {
+        fn to_tokens(&self, stream: &mut TokenStream) {
+            let module_name = match self.module_name {
+                ModuleName::Default => {
+                    let name_str = self.container_name.to_string().to_snake_case();
+                    syn::Ident::new(&name_str, self.container_name.span())
+                }
+                ModuleName::Custom(name) => name.clone(),
+            };
+
+            let visibility = self.visibility;
+            let body = self.body;
+
+            let module_tokens = quote! {
+                #visibility mod #module_name {
+                    use super::*;
+
+                    #body
+                }
+            };
+
+            stream.extend(module_tokens);
+        }
+    }
+}
+
+pub mod context_selector {
+    use crate::{ContextSelectorKind, Field, SuffixKind};
+    use proc_macro2::TokenStream;
+    use quote::{format_ident, quote, IdentFragment, ToTokens};
+
+    const DEFAULT_SUFFIX: &str = "Snafu";
 
     #[derive(Copy, Clone)]
     pub(crate) struct ContextSelector<'a> {
         pub backtrace_field: Option<&'a Field>,
+        pub implicit_fields: &'a [Field],
         pub crate_root: &'a dyn ToTokens,
         pub error_constructor_name: &'a dyn ToTokens,
         pub original_generics_without_defaults: &'a [TokenStream],
@@ -21,6 +69,7 @@ pub mod context_selector {
         pub user_fields: &'a [Field],
         pub visibility: Option<&'a dyn ToTokens>,
         pub where_clauses: &'a [TokenStream],
+        pub default_suffix: &'a SuffixKind,
     }
 
     impl ToTokens for ContextSelector<'_> {
@@ -43,6 +92,10 @@ pub mod context_selector {
                         #context_selector_into_error_impl
                     }
                 }
+                Whatever {
+                    source_field,
+                    message_field,
+                } => self.generate_whatever(source_field.as_ref(), message_field),
                 NoContext { source_field } => self.generate_from_source(source_field),
             };
 
@@ -65,8 +118,26 @@ pub mod context_selector {
         }
 
         fn parameterized_selector_name(&self) -> TokenStream {
-            let selector_name = self.selector_name;
+            let selector_name = self.selector_name.to_string();
+            let selector_name = selector_name.trim_end_matches("Error");
+            let suffix: &dyn IdentFragment = match self.selector_kind {
+                ContextSelectorKind::Context { suffix, .. } => {
+                    match suffix.resolve_with_default(self.default_suffix) {
+                        SuffixKind::Some(s) => s,
+                        SuffixKind::None => &"",
+                        SuffixKind::Default => &DEFAULT_SUFFIX,
+                    }
+                }
+                _ => &DEFAULT_SUFFIX,
+            };
+            let selector_name = format_ident!(
+                "{}{}",
+                selector_name,
+                suffix,
+                span = self.selector_name.span()
+            );
             let user_generics = self.user_field_generics();
+
             quote! { #selector_name<#(#user_generics,)*> }
         }
 
@@ -96,12 +167,16 @@ pub mod context_selector {
                 .collect()
         }
 
-        fn construct_backtrace_field(&self) -> Option<TokenStream> {
-            self.backtrace_field.map(|field| {
-                let crate_root = self.crate_root;
-                let name = &field.name;
-                quote! { #name: #crate_root::GenerateBacktrace::generate(), }
-            })
+        fn construct_implicit_fields(&self) -> TokenStream {
+            self.implicit_fields
+                .iter()
+                .chain(self.backtrace_field)
+                .map(|field| {
+                    let crate_root = self.crate_root;
+                    let name = &field.name;
+                    quote! { #name: #crate_root::GenerateImplicitData::generate(), }
+                })
+                .collect()
         }
 
         fn generate_type(self) -> TokenStream {
@@ -140,23 +215,27 @@ pub mod context_selector {
             let visibility = self.visibility;
             let extended_where_clauses = self.extended_where_clauses();
             let transfer_user_fields = self.transfer_user_fields();
-            let construct_backtrace_field = self.construct_backtrace_field();
+            let construct_implicit_fields = self.construct_implicit_fields();
+
+            let track_caller = track_caller();
 
             quote! {
                 impl<#(#user_field_generics,)*> #parameterized_selector_name {
                     #[doc = "Consume the selector and return the associated error"]
                     #[must_use]
+                    #track_caller
                     #visibility fn build<#(#original_generics_without_defaults,)*>(self) -> #parameterized_error_name
                     where
                         #(#extended_where_clauses),*
                     {
                         #error_constructor_name {
-                            #construct_backtrace_field
+                            #construct_implicit_fields
                             #(#transfer_user_fields,)*
                         }
                     }
 
                     #[doc = "Consume the selector and return a `Result` with the associated error"]
+                    #track_caller
                     #visibility fn fail<#(#original_generics_without_defaults,)* __T>(self) -> ::core::result::Result<__T, #parameterized_error_name>
                     where
                         #(#extended_where_clauses),*
@@ -176,7 +255,7 @@ pub mod context_selector {
             let user_field_generics = self.user_field_generics();
             let extended_where_clauses = self.extended_where_clauses();
             let transfer_user_fields = self.transfer_user_fields();
-            let construct_backtrace_field = self.construct_backtrace_field();
+            let construct_implicit_fields = self.construct_implicit_fields();
 
             let (source_ty, transfer_source_field) = match source_field {
                 Some(source_field) => {
@@ -186,6 +265,8 @@ pub mod context_selector {
                 None => (quote! { #crate_root::NoneError }, quote! {}),
             };
 
+            let track_caller = track_caller();
+
             quote! {
                 impl<#(#original_generics_without_defaults,)* #(#user_field_generics,)*> #crate_root::IntoError<#parameterized_error_name> for #parameterized_selector_name
                 where
@@ -194,10 +275,11 @@ pub mod context_selector {
                 {
                     type Source = #source_ty;
 
+                    #track_caller
                     fn into_error(self, error: Self::Source) -> #parameterized_error_name {
                         #error_constructor_name {
                             #transfer_source_field
-                            #construct_backtrace_field
+                            #construct_implicit_fields
                             #(#transfer_user_fields),*
                         }
                     }
@@ -205,25 +287,84 @@ pub mod context_selector {
             }
         }
 
+        fn generate_whatever(
+            self,
+            source_field: Option<&crate::SourceField>,
+            message_field: &crate::Field,
+        ) -> TokenStream {
+            let crate_root = self.crate_root;
+            let parameterized_error_name = self.parameterized_error_name;
+            let error_constructor_name = self.error_constructor_name;
+            let construct_implicit_fields = self.construct_implicit_fields();
+
+            // testme: transform
+
+            let (source_ty, transfer_source_field, empty_source_field) = match source_field {
+                Some(f) => {
+                    let source_field_type = f.transformation.ty();
+                    let source_field_name = &f.name;
+                    let source_transformation = f.transformation.transformation();
+
+                    (
+                        quote! { #source_field_type },
+                        Some(quote! { #source_field_name: (#source_transformation)(error), }),
+                        Some(quote! { #source_field_name: core::option::Option::None, }),
+                    )
+                }
+                None => (quote! { #crate_root::NoneError }, None, None),
+            };
+
+            let message_field_name = &message_field.name;
+
+            let track_caller = track_caller();
+
+            quote! {
+                impl #crate_root::FromString for #parameterized_error_name {
+                    type Source = #source_ty;
+
+                    #track_caller
+                    fn without_source(message: String) -> Self {
+                        #error_constructor_name {
+                            #empty_source_field
+                            #message_field_name: message,
+                            #construct_implicit_fields
+                        }
+                    }
+
+                    #track_caller
+                    fn with_source(error: Self::Source, message: String) -> Self {
+                        #error_constructor_name {
+                            #transfer_source_field
+                            #message_field_name: message,
+                            #construct_implicit_fields
+                        }
+                    }
+                }
+            }
+        }
+
         fn generate_from_source(self, source_field: &crate::SourceField) -> TokenStream {
             let parameterized_error_name = self.parameterized_error_name;
             let error_constructor_name = self.error_constructor_name;
-            let construct_backtrace_field = self.construct_backtrace_field();
+            let construct_implicit_fields = self.construct_implicit_fields();
             let original_generics_without_defaults = self.original_generics_without_defaults;
             let user_field_generics = self.user_field_generics();
             let where_clauses = self.where_clauses;
 
             let (source_field_type, transfer_source_field) = build_source_info(source_field);
 
+            let track_caller = track_caller();
+
             quote! {
                 impl<#(#original_generics_without_defaults,)* #(#user_field_generics,)*> ::core::convert::From<#source_field_type> for #parameterized_error_name
                 where
                     #(#where_clauses),*
                 {
+                    #track_caller
                     fn from(error: #source_field_type) -> Self {
                         #error_constructor_name {
                             #transfer_source_field
-                            #construct_backtrace_field
+                            #construct_implicit_fields
                         }
                     }
                 }
@@ -242,12 +383,21 @@ pub mod context_selector {
             quote! { #source_field_name: (#source_transformation)(error), },
         )
     }
+
+    fn track_caller() -> proc_macro2::TokenStream {
+        if cfg!(feature = "rust_1_46") {
+            quote::quote! { #[track_caller] }
+        } else {
+            quote::quote! {}
+        }
+    }
 }
 
 pub mod display {
     use crate::{Field, SourceField};
     use proc_macro2::TokenStream;
     use quote::{quote, ToTokens};
+    use std::collections::BTreeSet;
 
     struct StaticIdent(&'static str);
 
@@ -296,9 +446,10 @@ pub mod display {
 
     pub(crate) struct DisplayMatchArm<'a> {
         pub(crate) backtrace_field: Option<&'a crate::Field>,
+        pub(crate) implicit_fields: &'a [crate::Field],
         pub(crate) default_name: &'a dyn ToTokens,
-        pub(crate) display_format: Option<&'a dyn ToTokens>,
-        pub(crate) doc_comment: &'a str,
+        pub(crate) display_format: Option<&'a crate::Display>,
+        pub(crate) doc_comment: Option<&'a crate::DocComment>,
         pub(crate) pattern_ident: &'a dyn ToTokens,
         pub(crate) selector_kind: &'a crate::ContextSelectorKind,
     }
@@ -307,6 +458,7 @@ pub mod display {
         fn to_tokens(&self, stream: &mut TokenStream) {
             let Self {
                 backtrace_field,
+                implicit_fields,
                 default_name,
                 display_format,
                 doc_comment,
@@ -316,30 +468,52 @@ pub mod display {
 
             let user_fields = selector_kind.user_fields();
             let source_field = selector_kind.source_field();
+            let message_field = selector_kind.message_field();
 
-            let format = match (display_format, source_field) {
-                (Some(v), _) => quote! { #v },
-                (None, _) if !doc_comment.is_empty() => {
-                    quote! { #doc_comment }
+            let mut shorthand_names = &BTreeSet::new();
+            let mut assigned_names = &BTreeSet::new();
+
+            let format = match (display_format, doc_comment, source_field) {
+                (Some(v), _, _) => {
+                    let exprs = &v.exprs;
+                    shorthand_names = &v.shorthand_names;
+                    assigned_names = &v.assigned_names;
+                    quote! { #(#exprs),* }
                 }
-                (None, Some(f)) => {
+                (_, Some(d), _) => {
+                    let content = &d.content;
+                    shorthand_names = &d.shorthand_names;
+                    quote! { #content }
+                }
+                (_, _, Some(f)) => {
                     let field_name = &f.name;
                     quote! { concat!(stringify!(#default_name), ": {}"), #field_name }
                 }
-                (None, None) => quote! { stringify!(#default_name)},
+                _ => quote! { stringify!(#default_name)},
             };
 
             let field_names = user_fields
                 .iter()
                 .chain(backtrace_field)
+                .chain(implicit_fields)
+                .chain(message_field)
                 .map(Field::name)
-                .chain(source_field.map(SourceField::name));
+                .chain(source_field.map(SourceField::name))
+                .collect::<BTreeSet<_>>();
 
-            let field_names = quote! { #(ref #field_names),* };
+            let field_names_pat = quote! { #(ref #field_names),* };
+
+            let shorthand_names = shorthand_names.iter().collect::<BTreeSet<_>>();
+            let assigned_names = assigned_names.iter().collect::<BTreeSet<_>>();
+
+            let shorthand_fields = &shorthand_names & &field_names;
+            let shorthand_fields = &shorthand_fields - &assigned_names;
+
+            let shorthand_assignments = quote! { #( #shorthand_fields = #shorthand_fields ),* };
 
             let match_arm = quote! {
-                #pattern_ident { #field_names } => {
-                    write!(#FORMATTER_ARG, #format)
+                #pattern_ident { #field_names_pat } => {
+                    write!(#FORMATTER_ARG, #format, #shorthand_assignments)
                 }
             };
 
@@ -447,9 +621,20 @@ pub mod error {
                     let SourceField {
                         name: field_name, ..
                     } = source_field;
+
+                    let convert_to_error_source = if selector_kind.is_whatever() {
+                        quote! {
+                            #field_name.as_ref().map(|e| e.as_error_source())
+                        }
+                    } else {
+                        quote! {
+                            ::core::option::Option::Some(#field_name.as_error_source())
+                        }
+                    };
+
                     quote! {
                         #pattern_ident { ref #field_name, .. } => {
-                            ::core::option::Option::Some(#field_name.as_error_source())
+                            #convert_to_error_source
                         }
                     }
                 }
@@ -543,7 +728,7 @@ pub mod error_compat {
                         name: field_name, ..
                     } = backtrace_field;
                     quote! {
-                        #pattern_ident { ref #field_name, .. } => { #crate_root::GenerateBacktrace::as_backtrace(#field_name) }
+                        #pattern_ident { ref #field_name, .. } => { #crate_root::AsBacktrace::as_backtrace(#field_name) }
                     }
                 }
                 _ => {
